/*
Copyright 2017-2022 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package internal

import (
	"database/sql"
	"embed"
	"errors"
	"fmt"
	"log"

	"github.com/golang-migrate/migrate/v4"
	"github.com/golang-migrate/migrate/v4/database/sqlite3"
	"github.com/johejo/golang-migrate-extra/source/iofs"
)

//go:embed "migrations"
var Migrations embed.FS

type migrateLogger struct {
	log     *log.Logger
	verbose bool
}

func (m migrateLogger) Printf(format string, v ...interface{}) {
	m.log.Printf(format, v...)
}

func (m migrateLogger) Verbose() bool {
	return m.verbose
}

func InitializeDb(db *sql.DB, log *log.Logger) error {
	source, err := iofs.New(Migrations, "migrations")
	if err != nil {
		return fmt.Errorf("could not create migration source: %w", err)
	}

	driver, err := sqlite3.WithInstance(db, &sqlite3.Config{})
	if err != nil {
		return fmt.Errorf("could not create migration driver: %w", err)
	}

	m, err := migrate.NewWithInstance("iofs", source, "sqlite3", driver)
	if err != nil {
		return fmt.Errorf("could not create migration instance: %w", err)
	}

	m.Log = &migrateLogger{log, true}

	err = m.Up()
	if err != nil {
		if !errors.Is(err, migrate.ErrNoChange) {
			return fmt.Errorf("running database migration failed: %w", err)
		}

		log.Print("no database migrations required")
	} else {
		log.Print("applied database migrations")
	}

	return nil
}
